//------------------------------------------------------------------------------
// pv_app performance visualization application
//
// Copyright (c) 2014-2016 RWTH Aachen University, Germany,
// Virtual Reality & Immersive Visualisation Group.
//------------------------------------------------------------------------------
//                                License
//
// This framework is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// In the future, we may decide to add a commercial license
// at our own discretion without further notice.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

#include <cmath>

#include <QPainter>

#include <qpvt/delegates/directed_variation_glyph_delegate.hpp>
#include <qpvt/models/cnode_tree_model.hpp>

namespace qpvt {

//------------------------------------------------------------------------------
directed_variation_glyph_delegate::directed_variation_glyph_delegate
(QObject* p_parent)
: QAbstractItemDelegate(p_parent)
{
  
}


//------------------------------------------------------------------------------
void
directed_variation_glyph_delegate::paint
(QPainter*                   p_painter,
 const QStyleOptionViewItem& option,
 const QModelIndex&          index)
const
{
  const int left     = option.rect.x     ();
  const int top      = option.rect.y     ();
  const int height   = option.rect.height();
  const int baseline = top + height - 1;
  
  const QColor baseline_color(255 - 64, 255 - 64, 255 - 64);
  const QColor bar_color     (255 - 64, 255 - 64, 255 - 64);
  
  const QList<QVariant> directed_variations(
    index.data(cnode_tree_model::directed_variance_role).toList());

  const std::size_t num_dims = directed_variations.size();
  
  const double relative_severity =
    index.data(cnode_tree_model::get_relative_severity_role).toDouble();

  /// \todo make this user-defineable
  const double scaled_relative_severity =
    std::max(0.0, std::min(1.0, (relative_severity - 0.001) / 0.009));

  p_painter->save();
  
  // Clear painting area.
  p_painter->setPen  (Qt::NoPen);
  p_painter->setBrush(Qt::NoBrush);
  p_painter->drawRect(option.rect.x(),
                      option.rect.y(),
                      num_dims * option.rect.height(),
                      num_dims * option.rect.height());
  
  p_painter->setOpacity(m_fade_out_enabled ? scaled_relative_severity : 1.0);
  
  // Draw baseline.
  p_painter->setPen(baseline_color);
  for (std::size_t i = 0; i < num_dims; ++i)
    p_painter->drawLine(left + i * height + 1         , baseline,
                        left + i * height + height - 2, baseline);

  // Draw bars.
  p_painter->setBrush(bar_color);

  double max_dir_var = 1.0;
  for (std::size_t i = 0; i < num_dims; ++i)
    max_dir_var = std::max(max_dir_var, directed_variations[i].toDouble());
  
  for (std::size_t i = 0; i < num_dims; ++i)
  {
    const double curr_dir_var    = directed_variations[i].toDouble();
    const double clamped_dir_var = std::max(0.0, curr_dir_var);
    const double log_dir_var     = 
      log(clamped_dir_var / max_dir_var) / log(10.0);
    const double clamped_log_dir_var =
      std::max(0.0, std::min(1.0, (log_dir_var + 3.0) / 3.0));
    
    const double mapped_dir_var         = (clamped_log_dir_var - 0.001) / 0.999;
    const double clamped_mapped_dir_var =
      std::max(0.0, std::min(1.0, mapped_dir_var));
    
    const int max_height = height - 2;
    const int bar_height = clamped_mapped_dir_var * max_height;
    
    p_painter->drawRect(left + i * height + 1,
                        top  + 1 + max_height - bar_height,
                        height - 2,
                        bar_height);
  }
  
  p_painter->restore();
}


//------------------------------------------------------------------------------
QSize
directed_variation_glyph_delegate::sizeHint
(const QStyleOptionViewItem& option,
 const QModelIndex&          index)
const
{
  return QSize(4,4);
}

} // namespace qpvt
